#ifndef VEC4_H
#define VEC4_H

#include "General.h"

namespace SANC {

class Vec4 {

public:

  // Constructors.
  Vec4(double a_x = 0., double a_y = 0., double a_z = 0., double a_t = 0.)
    : m_x(a_x), m_y(a_y), m_z(a_z), m_t(a_t), m_s(M_NAN), b_s(false) { }
  Vec4(double a_x, double a_y, double a_z, double a_t, double a_s)
    : m_x(a_x), m_y(a_y), m_z(a_z), m_t(a_t), m_s(a_s), b_s(true) { }
  Vec4(const Vec4& v) : m_x(v.m_x), m_y(v.m_y), m_z(v.m_z), m_t(v.m_t), m_s(v.m_s), b_s(v.b_s) { }
//  Vec4(const Vec3& P, double E) : m_x(P.px()), m_y(P.py()), m_z(P.pz()), m_t(E), m_s(M_NAN), b_s(false) { }
  Vec4& operator=(const Vec4& v) { if (this != &v) { m_x = v.m_x; m_y = v.m_y;
    m_z = v.m_z; m_t = v.m_t; m_s = v.m_s; b_s = v.b_s;} return *this; }
  Vec4& operator=(double value) { m_x = value; m_y = value; m_z = value;
    m_t = value; m_s = M_NAN; b_s = false; return *this; }

  // Member functions for input.
  void reset() {m_x = 0.; m_y = 0.; m_z = 0.; m_t = 0.; m_s = 0.; b_s = false;}
  void p(double a_x, double a_y, double a_z, double a_t)
    {m_x = a_x; m_y = a_y; m_z = a_z; m_t = a_t;}
  void p(Vec4 a_p) {m_x = a_p.m_x; m_y = a_p.m_y; m_z = a_p.m_z; m_t = a_p.m_t; m_s = a_p.m_s; b_s = a_p.b_s;}
  void px(double a_x) {m_x = a_x;}
  void py(double a_y) {m_y = a_y;}
  void pz(double a_z) {m_z = a_z;}
  void e(double a_t) {m_t = a_t;}
  void s(double a_s) {m_s = a_s; b_s = true;}

  // Member functions for output.
  double px() const {return m_x;}
  double py() const {return m_y;}
  double pz() const {return m_z;}
  double e() const {return m_t;}
  double s() const {return b_s?m_s:mCalc();}
  bool isS() const {return b_s;}
  double mCalc() const {double temp = m_t*m_t - m_x*m_x - m_y*m_y - m_z*m_z;
    return (temp >= 0.) ? sqrt(temp) : -sqrt(-temp);}
  double m2Calc() const {return m_t*m_t - m_x*m_x - m_y*m_y - m_z*m_z;}
  double pT() const {return sqrt(m_x*m_x + m_y*m_y);}
  double pT2() const {return m_x*m_x + m_y*m_y;}
  double pAbs() const {return sqrt(m_x*m_x + m_y*m_y + m_z*m_z);}
  double pAbs2() const {return m_x*m_x + m_y*m_y + m_z*m_z;}
  double eT() const {double temp = m_x*m_x + m_y*m_y;
    return m_t * sqrt( temp / (temp + m_z*m_z) );}
  double eT2() const {double temp = m_x*m_x + m_y*m_y;
    return m_t*m_t * temp / (temp + m_z*m_z);}
  double theta() const {return atan2(sqrt(m_x*m_x + m_y*m_y), m_z);}
  double phi() const {return atan2(m_y,m_x);}
  double thetaXZ() const {return atan2(m_x,m_z);}
  double pPos() const {return m_t + m_z;}
  double pNeg() const {return m_t - m_z;}
  double rap() const {return 0.5 * log( (m_t + m_z) / (m_t - m_z) );}
  double eta() const {double xyz = sqrt(m_x*m_x + m_y*m_y + m_z*m_z);
    return 0.5 * log( (xyz + m_z) / (xyz - m_z) );}

  // Member functions that perform operations.
  Vec4& rescale3(double fac) {m_x *= fac; m_y *= fac; m_z *= fac; return *this;}
  Vec4& rescale4(double fac) {m_x *= fac; m_y *= fac; m_z *= fac; m_t *= fac; return *this;}
  Vec4& flipX() {m_x = -m_x; return *this;}
  Vec4& flipY() {m_y = -m_y; return *this;}
  Vec4& flipZ() {m_z = -m_z; return *this;}
  Vec4& flip3() {m_x = -m_x; m_y = -m_y; m_z = -m_z; return *this;}
  Vec4& flip4() {m_x = -m_x; m_y = -m_y; m_z = -m_z; m_t = -m_t; return *this;}
  Vec4& rot(double sth, double cth, double sph, double cph);
  Vec4& bstZ(double betaZ);

  // Operator overloading with member functions
  Vec4 operator-()const {Vec4 tmp; tmp.m_x = -m_x; tmp.m_y = -m_y; tmp.m_z = -m_z;
    tmp.m_t = -m_t; tmp.m_s = m_s; tmp.b_s = b_s; return tmp;}
  Vec4& operator+=(const Vec4& v) {m_x += v.m_x; m_y += v.m_y; m_z += v.m_z;
    m_t += v.m_t; m_s = M_NAN; b_s = false; return *this;}
  Vec4& operator-=(const Vec4& v) {m_x -= v.m_x; m_y -= v.m_y; m_z -= v.m_z;
    m_t -= v.m_t; m_s = M_NAN; b_s = false; return *this;}
  Vec4& operator*=(double f) {m_x *= f; m_y *= f; m_z *= f;
    m_t *= f; m_s = M_NAN; b_s = false; return *this;}
  Vec4& operator/=(double f) {m_x /= f; m_y /= f; m_z /= f;
    m_t /= f; m_s = M_NAN; b_s = false; return *this;}

  // Operator overloading with friends
  friend Vec4 operator+(const Vec4& v1, const Vec4& v2);
  friend Vec4 operator-(const Vec4& v1, const Vec4& v2);
  friend Vec4 operator*(double f, const Vec4& v1);
  friend Vec4 operator*(const Vec4& v1, double f);
  friend Vec4 operator/(const Vec4& v1, double f);
  friend double operator*(const Vec4& v1, const Vec4& v2);

  // Invariant mass of a pair and its square.
  friend double m(const Vec4& v1, const Vec4& v2);
  friend double m2(const Vec4& v1, const Vec4& v2);

  // Scalar and cross product of 3-vector parts.
  friend double dot3(const Vec4& v1, const Vec4& v2);
  friend Vec4 cross3(const Vec4& v1, const Vec4& v2);

  // theta is polar angle between v1 and v2.
  friend double theta(const Vec4& v1, const Vec4& v2);
  friend double costheta(const Vec4& v1, const Vec4& v2);

  // phi is azimuthal angle between v1 and v2 around z axis.
  friend double phi(const Vec4& v1, const Vec4& v2);
  friend double cosphi(const Vec4& v1, const Vec4& v2);

  // phi is azimuthal angle between v1 and v2 around n axis.
  friend double phi(const Vec4& v1, const Vec4& v2, const Vec4& n);
  friend double cosphi(const Vec4& v1, const Vec4& v2, const Vec4& n);

  // R is distance in cylindrical (y/eta, phi) coordinates.
  friend double RRapPhi(const Vec4& v1, const Vec4& v2);
  friend double REtaPhi(const Vec4& v1, const Vec4& v2);

  // Print a four-vector.
  friend std::ostream& operator<<(std::ostream&, const Vec4& v) ;

private:
  // The four-vector data members.
  double m_x, m_y, m_z, m_t, m_s;

  bool b_s;

  static const double reg;
};

// Namespace function declarations; friends of Vec4 class.

// Implementation of operator overloading with friends.

inline Vec4 operator+(const Vec4& v1, const Vec4& v2)
  {Vec4 v = v1 ; return v += v2;}

inline Vec4 operator-(const Vec4& v1, const Vec4& v2)
  {Vec4 v = v1 ; return v -= v2;}

inline Vec4 operator*(double f, const Vec4& v1)
  {Vec4 v = v1; return v *= f;}

inline Vec4 operator*(const Vec4& v1, double f)
  {Vec4 v = v1; return v *= f;}

inline Vec4 operator/(const Vec4& v1, double f)
  {Vec4 v = v1; return v /= f;}

inline double operator*(const Vec4& v1, const Vec4& v2)
  {return v1.m_t*v2.m_t - v1.m_x*v2.m_x - v1.m_y*v2.m_y - v1.m_z*v2.m_z;}

// Invariant mass of a pair and its square.
double m(const Vec4& v1, const Vec4& v2);
double m2(const Vec4& v1, const Vec4& v2);

// Scalar and cross product of 3-vector parts.
double dot3(const Vec4& v1, const Vec4& v2);
Vec4 cross3(const Vec4& v1, const Vec4& v2);

// theta is polar angle between v1 and v2.
double theta(const Vec4& v1, const Vec4& v2);
double costheta(const Vec4& v1, const Vec4& v2);

// phi is azimuthal angle between v1 and v2 around z axis.
double phi(const Vec4& v1, const Vec4& v2);
double cosphi(const Vec4& v1, const Vec4& v2);

// phi is azimuthal angle between v1 and v2 around n axis.
double phi(const Vec4& v1, const Vec4& v2, const Vec4& n);
double cosphi(const Vec4& v1, const Vec4& v2, const Vec4& n);

// R is distance in cylindrical (y/eta, phi) coordinates.
double RRapPhi(const Vec4& v1, const Vec4& v2);
double REtaPhi(const Vec4& v1, const Vec4& v2);

// Print a four-vector.
std::ostream& operator<<(std::ostream&, const Vec4& v) ;

inline double ENorm2(Vec4 v){
    return pow2(v.e())+v.pAbs2();
}
inline double ENorm(Vec4 v){
    return sqrt(ENorm2(v));
}

inline double UNorm(Vec4 v){
    return fabs(v.e())+fabs(v.px())+fabs(v.py())+fabs(v.pz());
}

inline double pow2(Vec4 x){return x.s();}

}

#endif // VEC4_H
